@@ -43,7 +43,6 @@ import subprocess
 import zlib
 from base64 import a85decode
 from dataclasses import dataclass
-from io import BytesIO
 from pathlib import Path
 from tempfile import TemporaryDirectory
 from typing import Any, Optional, Union, cast
@@ -51,6 +50,7 @@ from typing import Any, Optional, Union, cast
 from ._codecs._codecs import LzwCodec as _LzwCodec
 from ._utils import (
     WHITESPACES_AS_BYTES,
+    deprecate,
     deprecation_with_replacement,
     logger_warning,
 )
@@ -66,12 +66,19 @@ from .generic import (
     DictionaryObject,
     IndirectObject,
     NullObject,
+    NumberObject,
     StreamObject,
     is_null_or_none,
 )
 
+JBIG2_MAX_OUTPUT_LENGTH = 75_000_000
+LZW_MAX_OUTPUT_LENGTH = 75_000_000
+RUN_LENGTH_MAX_OUTPUT_LENGTH = 75_000_000
 ZLIB_MAX_OUTPUT_LENGTH = 75_000_000
-LZW_MAX_OUTPUT_LENGTH = 1_000_000_000
+ZLIB_MAX_RECOVERY_INPUT_LENGTH = 5_000_000
+
+# Reuse cached 1-byte values in the fallback loop to avoid per-byte allocations.
+_SINGLE_BYTES = tuple(bytes((i,)) for i in range(256))
 
 
 def _decompress_with_limit(data: bytes) -> bytes:
@@ -130,18 +137,28 @@ def decompress(data: bytes) -> bytes:
         decompressor = zlib.decompressobj(zlib.MAX_WBITS | 32)
         result_str = b""
         remaining_limit = ZLIB_MAX_OUTPUT_LENGTH
-        data_single_bytes = [data[i : i + 1] for i in range(len(data))]
-        for index, b in enumerate(data_single_bytes):
+        data_length = len(data)
+        known_errors = set()
+        for index in range(data_length):
+            chunk = _SINGLE_BYTES[data[index]]
             try:
-                decompressed = decompressor.decompress(b, max_length=remaining_limit)
+                decompressed = decompressor.decompress(chunk, max_length=remaining_limit)
                 result_str += decompressed
                 remaining_limit -= len(decompressed)
                 if remaining_limit <= 0:
                     raise LimitReachedError(
-                        f"Limit reached while decompressing. {len(data_single_bytes) - index} bytes remaining."
+                        f"Limit reached while decompressing. {data_length - index} bytes remaining."
                     )
-            except zlib.error:
-                pass
+            except zlib.error as error:
+                if index > ZLIB_MAX_RECOVERY_INPUT_LENGTH:
+                    raise LimitReachedError(
+                        f"Recovery limit reached while decompressing. {data_length - index} bytes remaining."
+                    )
+                error_str = str(error)
+                if error_str in known_errors:
+                    continue
+                logger_warning(error_str, __name__)
+                known_errors.add(error_str)
         return result_str
 
 
@@ -156,45 +173,28 @@ class FlateDecode:
         Decode data which is flate-encoded.
 
         Args:
-          data: flate-encoded data.
-          decode_parms: a dictionary of values, understanding the
-            "/Predictor":<int> key only
+          data: Flate-encoded data.
+          decode_parms: Additional decoding parameters.
 
         Returns:
           The flate-decoded data.
 
         Raises:
-          PdfReadError:
+          PdfReadError: Unsupported parameters have been found.
 
         """
         str_data = decompress(data)
-        predictor = 1
 
-        if decode_parms:
-            try:
-                predictor = decode_parms.get("/Predictor", 1)
-            except (AttributeError, TypeError):  # Type Error is NullObject
-                pass  # Usually an array with a null object was read
+        if isinstance(decode_parms, DictionaryObject):
+            parameters = decode_parms
+        else:
+            parameters = DictionaryObject()
+
+        predictor = parameters.get("/Predictor", 1)
+
         # predictor 1 == no predictor
         if predictor != 1:
-            # /Columns, the number of samples in each row, has a default value of 1;
-            # §7.4.4.3, ISO 32000.
-            DEFAULT_BITS_PER_COMPONENT = 8
-            try:
-                columns = cast(int, decode_parms[LZW.COLUMNS].get_object())  # type: ignore
-            except (TypeError, KeyError):
-                columns = 1
-            try:
-                colors = cast(int, decode_parms[LZW.COLORS].get_object())  # type: ignore
-            except (TypeError, KeyError):
-                colors = 1
-            try:
-                bits_per_component = cast(
-                    int,
-                    decode_parms[LZW.BITS_PER_COMPONENT].get_object(),  # type: ignore
-                )
-            except (TypeError, KeyError):
-                bits_per_component = DEFAULT_BITS_PER_COMPONENT
+            columns, colors, bits_per_component = FlateDecode._get_parameters(parameters)
 
             # PNG predictor can vary by row and so is the lead byte on each row
             rowlength = (
@@ -219,6 +219,20 @@ class FlateDecode:
                 raise PdfReadError(f"Unsupported flatedecode predictor {predictor!r}")
         return str_data
 
+    @staticmethod
+    def _get_parameters(parameters: DictionaryObject) -> tuple[int, int, int]:
+        # For details, see table 8 of ISO 32000-2:2020.
+        def get(key: str, default: int) -> int:
+            _value = parameters.get(key, NumberObject(default)).get_object()
+            if not isinstance(_value, int) or _value < 1:
+                raise PdfReadError(f"Expected positive number for {key}, got {_value}!")
+            return _value
+
+        columns = get(key=LZW.COLUMNS, default=1)
+        colors = get(key=LZW.COLORS, default=1)
+        bits_per_component = get(key=LZW.BITS_PER_COMPONENT, default=8)
+        return columns, colors, bits_per_component
+
     @staticmethod
     def _decode_png_prediction(data: bytes, columns: int, rowlength: int) -> bytes:
         # PNG prediction can vary from row to row
@@ -395,8 +409,10 @@ class RunLengthDecode:
         """
         lst = []
         index = 0
+        data_length = len(data)
+        total_length = 0
         while True:
-            if index >= len(data):
+            if index >= data_length:
                 logger_warning(
                     "missing EOD in RunLengthDecode, check if output is OK", __name__
                 )
@@ -404,17 +420,21 @@ class RunLengthDecode:
             length = data[index]
             index += 1
             if length == 128:
-                data_length = len(data)
                 if index < data_length:
                     # We should first check, if we have an inner stream from a multi-encoded
                     # stream with a faulty trailing newline that we can decode properly.
                     # We will just ignore the last byte and raise a warning ...
-                    if (index == data_length - 1) and (data[index : index+1] == b"\n"):
+                    if (index == data_length - 1) and (data[index : index + 1] == b"\n"):
                         logger_warning(
                             "Found trailing newline in stream data, check if output is OK", __name__
                         )
                         break
-                    raise PdfStreamError("Early EOD in RunLengthDecode")
+                    # Raising an exception here breaks all image extraction for this file, which might
+                    # not be desirable. For this reason, indicate that the output is most likely wrong,
+                    # as processing stopped after the first EOD marker. See issue #3517.
+                    logger_warning(
+                        "Early EOD in RunLengthDecode, check if output is OK", __name__
+                    )
                 break
             if length < 128:
                 length += 1
@@ -424,6 +444,9 @@ class RunLengthDecode:
                 length = 257 - length
                 lst.append(bytes((data[index],)) * length)
                 index += 1
+            total_length += length
+            if total_length > RUN_LENGTH_MAX_OUTPUT_LENGTH:
+                raise LimitReachedError("Limit reached while decompressing.")
         return b"".join(lst)
 
 
@@ -482,6 +505,8 @@ class ASCII85Decode:
         if isinstance(data, str):
             data = data.encode()
         data = data.strip(WHITESPACES_AS_BYTES)
+        if len(data) > 2 and data.endswith(b">"):
+            data = data[:-1].rstrip(WHITESPACES_AS_BYTES) + data[-1:]
         try:
             return a85decode(data, adobe=True, ignorechars=WHITESPACES_AS_BYTES)
         except ValueError as error:
@@ -701,12 +726,23 @@ class JBIG2Decode:
             environment = os.environ.copy()
             environment["LC_ALL"] = "C"
             result = subprocess.run(  # noqa: S603
-                [JBIG2DEC_BINARY, "--embedded", "--format", "png", "--output", "-", *paths],
+                [
+                    JBIG2DEC_BINARY,
+                    "--embedded",
+                    "--format", "png",
+                    "--output", "-",
+                    "-M", str(JBIG2_MAX_OUTPUT_LENGTH),
+                    *paths
+                ],
                 capture_output=True,
                 env=environment,
             )
-            if b"unrecognized option '--embedded'" in result.stderr:
-                raise DependencyError("jbig2dec>=0.15 is required.")
+            if b"unrecognized option '--embedded'" in result.stderr or b"unrecognized option '-M'" in result.stderr:
+                raise DependencyError("jbig2dec>=0.19 is required.")
+            if b"FATAL ERROR failed to allocate image data buffer" in result.stderr:
+                raise LimitReachedError(
+                    f"Memory limit reached while reading JBIG2 data:\n{result.stderr.decode('utf-8')}"
+                )
             if result.stderr:
                 for line in result.stderr.decode("utf-8").splitlines():
                     logger_warning(line, __name__)
@@ -726,10 +762,19 @@ class JBIG2Decode:
         version = result.stdout.split(" ", maxsplit=1)[1]
 
         from ._utils import Version  # noqa: PLC0415
-        return Version(version) >= Version("0.15")
+        return Version(version) >= Version("0.19")
 
 
-def decode_stream_data(stream: Any) -> bytes:
+def _deprecate_inline_image_filters(filter_name: str, old_name: str, new_name: str) -> None:
+    if filter_name != old_name:
+        return
+    deprecate(
+        f"The filter name {old_name} is deprecated and will be removed in pypdf 7.0.0. Use {new_name} instead.",
+        4,
+    )
+
+
+def decode_stream_data(stream: StreamObject) -> bytes:
     """
     Decode the stream data based on the specified filters.
 
@@ -763,19 +808,26 @@ def decode_stream_data(stream: Any) -> bytes:
         if isinstance(params, NullObject):
             params = {}
         if filter_name in (FT.ASCII_HEX_DECODE, FTA.AHx):
+            _deprecate_inline_image_filters(filter_name=filter_name, old_name=FTA.AHx, new_name=FT.ASCII_HEX_DECODE)
             data = ASCIIHexDecode.decode(data)
         elif filter_name in (FT.ASCII_85_DECODE, FTA.A85):
+            _deprecate_inline_image_filters(filter_name=filter_name, old_name=FTA.A85, new_name=FT.ASCII_85_DECODE)
             data = ASCII85Decode.decode(data)
         elif filter_name in (FT.LZW_DECODE, FTA.LZW):
+            _deprecate_inline_image_filters(filter_name=filter_name, old_name=FTA.LZW, new_name=FT.LZW_DECODE)
             data = LZWDecode.decode(data, params)
         elif filter_name in (FT.FLATE_DECODE, FTA.FL):
+            _deprecate_inline_image_filters(filter_name=filter_name, old_name=FTA.FL, new_name=FT.FLATE_DECODE)
             data = FlateDecode.decode(data, params)
         elif filter_name in (FT.RUN_LENGTH_DECODE, FTA.RL):
+            _deprecate_inline_image_filters(filter_name=filter_name, old_name=FTA.RL, new_name=FT.RUN_LENGTH_DECODE)
             data = RunLengthDecode.decode(data)
-        elif filter_name == FT.CCITT_FAX_DECODE:
+        elif filter_name in (FT.CCITT_FAX_DECODE, FTA.CCF):
+            _deprecate_inline_image_filters(filter_name=filter_name, old_name=FTA.CCF, new_name=FT.CCITT_FAX_DECODE)
             height = stream.get(IA.HEIGHT, ())
             data = CCITTFaxDecode.decode(data, params, height)
-        elif filter_name == FT.DCT_DECODE:
+        elif filter_name in (FT.DCT_DECODE, FTA.DCT):
+            _deprecate_inline_image_filters(filter_name=filter_name, old_name=FTA.DCT, new_name=FT.DCT_DECODE)
             data = DCTDecode.decode(data)
         elif filter_name == FT.JPX_DECODE:
             data = JPXDecode.decode(data)
@@ -789,195 +841,3 @@ def decode_stream_data(stream: Any) -> bytes:
         else:
             raise NotImplementedError(f"Unsupported filter {filter_name}")
     return data
-
-
-def _xobj_to_image(
-        x_object: dict[str, Any],
-        pillow_parameters: Union[dict[str, Any], None] = None
-) -> tuple[Optional[str], bytes, Any]:
-    """
-    Users need to have the pillow package installed.
-
-    It's unclear if pypdf will keep this function here, hence it's private.
-    It might get removed at any point.
-
-    Args:
-        x_object:
-        pillow_parameters: parameters provided to Pillow Image.save() method,
-            cf. <https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.save>
-
-    Returns:
-        Tuple[file extension, bytes, PIL.Image.Image]
-
-    """
-    from ._xobj_image_helpers import (  # noqa: PLC0415
-        Image,
-        UnidentifiedImageError,
-        _apply_decode,
-        _extended_image_frombytes,
-        _get_mode_and_invert_color,
-        _handle_flate,
-        _handle_jpx,
-    )
-
-    def _apply_alpha(
-        img: Image.Image,
-        x_object: dict[str, Any],
-        obj_as_text: str,
-        image_format: str,
-        extension: str,
-    ) -> tuple[Image.Image, str, str]:
-        alpha = None
-        if IA.S_MASK in x_object:  # add alpha channel
-            alpha = _xobj_to_image(x_object[IA.S_MASK])[2]
-            if img.size != alpha.size:
-                logger_warning(
-                    f"image and mask size not matching: {obj_as_text}", __name__
-                )
-            else:
-                # TODO: implement mask
-                if alpha.mode != "L":
-                    alpha = alpha.convert("L")
-                if img.mode == "P":
-                    img = img.convert("RGB")
-                elif img.mode == "1":
-                    img = img.convert("L")
-                img.putalpha(alpha)
-            if "JPEG" in image_format:
-                image_format = "JPEG2000"
-                extension = ".jp2"
-            else:
-                image_format = "PNG"
-                extension = ".png"
-        return img, extension, image_format
-
-    # For error reporting
-    obj_as_text = (
-        x_object.indirect_reference.__repr__()
-        if x_object is None  # pragma: no cover
-        else x_object.__repr__()
-    )
-
-    # Get size and data
-    size = (cast(int, x_object[IA.WIDTH]), cast(int, x_object[IA.HEIGHT]))
-    data = x_object.get_data()  # type: ignore
-    if isinstance(data, str):  # pragma: no cover
-        data = data.encode()
-    if len(data) % (size[0] * size[1]) == 1 and data[-1] == 0x0A:  # ie. '\n'
-        data = data[:-1]
-
-    # Get color properties
-    colors = x_object.get("/Colors", 1)
-    color_space: Any = x_object.get("/ColorSpace", NullObject()).get_object()
-    if isinstance(color_space, list) and len(color_space) == 1:
-        color_space = color_space[0].get_object()
-
-    mode, invert_color = _get_mode_and_invert_color(x_object, colors, color_space)
-
-    # Get filters
-    filters = x_object.get(SA.FILTER, NullObject()).get_object()
-    lfilters = filters[-1] if isinstance(filters, list) else filters
-    decode_parms = x_object.get(SA.DECODE_PARMS, None)
-    if decode_parms and isinstance(decode_parms, (tuple, list)):
-        decode_parms = decode_parms[0]
-    else:
-        decode_parms = {}
-    if not isinstance(decode_parms, dict):
-        decode_parms = {}
-
-    extension = None
-    if lfilters in (FT.FLATE_DECODE, FT.RUN_LENGTH_DECODE):
-        img, image_format, extension, _ = _handle_flate(
-            size,
-            data,
-            mode,
-            color_space,
-            colors,
-            obj_as_text,
-        )
-    elif lfilters in (FT.LZW_DECODE, FT.ASCII_85_DECODE):
-        # I'm not sure if the following logic is correct.
-        # There might not be any relationship between the filters and the
-        # extension
-        if lfilters == FT.LZW_DECODE:
-            image_format = "TIFF"
-            extension = ".tiff"  # mime_type = "image/tiff"
-        else:
-            image_format = "PNG"
-            extension = ".png"  # mime_type = "image/png"
-        try:
-            img = Image.open(BytesIO(data), formats=("TIFF", "PNG"))
-        except UnidentifiedImageError:
-            img = _extended_image_frombytes(mode, size, data)
-    elif lfilters == FT.DCT_DECODE:
-        img, image_format, extension = Image.open(BytesIO(data)), "JPEG", ".jpg"
-        # invert_color kept unchanged
-    elif lfilters == FT.JPX_DECODE:
-        img, image_format, extension, invert_color = _handle_jpx(
-            size, data, mode, color_space, colors
-        )
-    elif lfilters == FT.CCITT_FAX_DECODE:
-        img, image_format, extension, invert_color = (
-            Image.open(BytesIO(data), formats=("TIFF",)),
-            "TIFF",
-            ".tiff",
-            False,
-        )
-    elif lfilters == FT.JBIG2_DECODE:
-        img, image_format, extension, invert_color = (
-            Image.open(BytesIO(data), formats=("PNG",)),
-            "PNG",
-            ".png",
-            False,
-        )
-    elif mode == "CMYK":
-        img, image_format, extension, invert_color = (
-            _extended_image_frombytes(mode, size, data),
-            "TIFF",
-            ".tif",
-            False,
-        )
-    elif mode == "":
-        raise PdfReadError(f"ColorSpace field not found in {x_object}")
-    else:
-        img, image_format, extension, invert_color = (
-            _extended_image_frombytes(mode, size, data),
-            "PNG",
-            ".png",
-            False,
-        )
-
-    img = _apply_decode(img, x_object, lfilters, color_space, invert_color)
-    img, extension, image_format = _apply_alpha(
-        img, x_object, obj_as_text, image_format, extension
-    )
-
-    if pillow_parameters is None:
-        pillow_parameters = {}
-    # Preserve JPEG image quality - see issue #3515.
-    if image_format == "JPEG":
-        # This prevents: Cannot use 'keep' when original image is not a JPEG:
-        # "JPEG" is the value of PIL.JpegImagePlugin.JpegImageFile.format
-        img.format = "JPEG"  # type: ignore[misc]
-        if "quality" not in pillow_parameters:
-            pillow_parameters["quality"] = "keep"
-
-    # Save image to bytes
-    img_byte_arr = BytesIO()
-    try:
-        img.save(img_byte_arr, format=image_format, **pillow_parameters)
-    except OSError:  # pragma: no cover  # covered with pillow 10.3
-        # in case of we convert to RGBA and then to PNG
-        img1 = img.convert("RGBA")
-        image_format = "PNG"
-        extension = ".png"
-        img_byte_arr = BytesIO()
-        img1.save(img_byte_arr, format=image_format)
-    data = img_byte_arr.getvalue()
-
-    try:  # temporary try/except until other fixes of images
-        img = Image.open(BytesIO(data))
-    except Exception as exception:
-        logger_warning(f"Failed loading image: {exception}", __name__)
-        img = None  # type: ignore
-    return extension, data, img
