@@ -30,20 +30,28 @@ import hashlib
 import re
 import sys
 from binascii import unhexlify
+from collections.abc import Sequence
 from math import log10
 from struct import iter_unpack
-from typing import Any, Callable, ClassVar, Dict, Optional, Sequence, Union, cast
+from typing import Any, Callable, ClassVar, Optional, Union, cast
 
 if sys.version_info[:2] >= (3, 10):
     from typing import TypeGuard
 else:
     from typing_extensions import TypeGuard  # PEP 647
 
+if sys.version_info >= (3, 11):
+    from typing import Self
+else:
+    from typing_extensions import Self
+
 from .._codecs import _pdfdoc_encoding_rev
 from .._protocols import PdfObjectProtocol, PdfWriterProtocol
 from .._utils import (
     StreamType,
-    deprecate_no_replacement,
+    classproperty,
+    deprecation_no_replacement,
+    deprecation_with_replacement,
     logger_warning,
     read_non_whitespace,
     read_until_regex,
@@ -65,21 +73,19 @@ class PdfObject(PdfObjectProtocol):
 
         Returns:
             Hash considering type and value.
+
         """
         raise NotImplementedError(
             f"{self.__class__.__name__} does not implement .hash_bin() so far"
         )
 
     def hash_value_data(self) -> bytes:
-        return ("%s" % self).encode()
+        return f"{self}".encode()
 
     def hash_value(self) -> bytes:
         return (
-            "%s:%s"
-            % (
-                self.__class__.__name__,
-                self.hash_func(self.hash_value_data()).hexdigest(),
-            )
+            f"{self.__class__.__name__}:"
+            f"{self.hash_func(self.hash_value_data()).hexdigest()}"
         ).encode()
 
     def replicate(
@@ -95,6 +101,7 @@ class PdfObject(PdfObjectProtocol):
 
         Returns:
           The cloned PdfObject
+
         """
         return self.clone(pdf_dest)
 
@@ -123,6 +130,7 @@ class PdfObject(PdfObjectProtocol):
 
         Returns:
           The cloned PdfObject
+
         """
         raise NotImplementedError(
             f"{self.__class__.__name__} does not implement .clone so far"
@@ -143,6 +151,7 @@ class PdfObject(PdfObjectProtocol):
 
         Returns:
           The clone
+
         """
         try:
             if not force_duplicate and clone.indirect_reference.pdf == pdf_dest:
@@ -166,7 +175,7 @@ class PdfObject(PdfObjectProtocol):
         if ind is not None:
             if id(ind.pdf) not in pdf_dest._id_translated:
                 pdf_dest._id_translated[id(ind.pdf)] = {}
-                pdf_dest._id_translated[id(ind.pdf)]["PreventGC"] = ind.pdf  # type: ignore
+                pdf_dest._id_translated[id(ind.pdf)]["PreventGC"] = ind.pdf  # type: ignore[index]
             if (
                 not force_duplicate
                 and ind.idnum in pdf_dest._id_translated[id(ind.pdf)]
@@ -213,6 +222,7 @@ class NullObject(PdfObject):
 
         Returns:
             Hash considering type and value.
+
         """
         return hash((self.__class__,))
 
@@ -220,7 +230,7 @@ class NullObject(PdfObject):
         self, stream: StreamType, encryption_key: Union[None, str, bytes] = None
     ) -> None:
         if encryption_key is not None:  # deprecated
-            deprecate_no_replacement(
+            deprecation_no_replacement(
                 "the encryption_key parameter of write_to_stream", "5.0.0"
             )
         stream.write(b"null")
@@ -235,6 +245,12 @@ class NullObject(PdfObject):
     def __repr__(self) -> str:
         return "NullObject"
 
+    def __eq__(self, other: object) -> bool:
+        return isinstance(other, NullObject)
+
+    def __hash__(self) -> int:
+        return self.hash_bin()
+
 
 class BooleanObject(PdfObject):
     def __init__(self, value: Any) -> None:
@@ -258,16 +274,19 @@ class BooleanObject(PdfObject):
 
         Returns:
             Hash considering type and value.
+
         """
         return hash((self.__class__, self.value))
 
-    def __eq__(self, __o: object) -> bool:
-        if isinstance(__o, BooleanObject):
-            return self.value == __o.value
-        elif isinstance(__o, bool):
-            return self.value == __o
-        else:
-            return False
+    def __eq__(self, o: object, /) -> bool:
+        if isinstance(o, BooleanObject):
+            return self.value == o.value
+        if isinstance(o, bool):
+            return self.value == o
+        return False
+
+    def __hash__(self) -> int:
+        return self.hash_bin()
 
     def __repr__(self) -> str:
         return "True" if self.value else "False"
@@ -276,7 +295,7 @@ class BooleanObject(PdfObject):
         self, stream: StreamType, encryption_key: Union[None, str, bytes] = None
     ) -> None:
         if encryption_key is not None:  # deprecated
-            deprecate_no_replacement(
+            deprecation_no_replacement(
                 "the encryption_key parameter of write_to_stream", "5.0.0"
             )
         if self.value:
@@ -289,11 +308,10 @@ class BooleanObject(PdfObject):
         word = stream.read(4)
         if word == b"true":
             return BooleanObject(True)
-        elif word == b"fals":
+        if word == b"fals":
             stream.read(1)
             return BooleanObject(False)
-        else:
-            raise PdfReadError("Could not read Boolean object")
+        raise PdfReadError("Could not read Boolean object")
 
 
 class IndirectObject(PdfObject):
@@ -311,6 +329,7 @@ class IndirectObject(PdfObject):
 
         Returns:
             Hash considering type and value.
+
         """
         return hash((self.__class__, self.idnum, self.generation, id(self.pdf)))
 
@@ -332,6 +351,7 @@ class IndirectObject(PdfObject):
             return self
         if id(self.pdf) not in pdf_dest._id_translated:
             pdf_dest._id_translated[id(self.pdf)] = {}
+            pdf_dest._id_translated[id(self.pdf)]["PreventGC"] = self.pdf  # type: ignore[index]
 
         if self.idnum in pdf_dest._id_translated[id(self.pdf)]:
             dup = pdf_dest.get_object(pdf_dest._id_translated[id(self.pdf)][self.idnum])
@@ -351,9 +371,8 @@ class IndirectObject(PdfObject):
             dup = pdf_dest._add_object(
                 obj.clone(pdf_dest, force_duplicate, ignore_fields)
             )
-        # asserts added to prevent errors in mypy
-        assert dup is not None
-        assert dup.indirect_reference is not None
+        assert dup is not None, "mypy"
+        assert dup.indirect_reference is not None, "mypy"
         return dup.indirect_reference
 
     @property
@@ -388,10 +407,20 @@ class IndirectObject(PdfObject):
         # items should be extracted from pointed Object
         return self._get_object_with_check()[key]  # type: ignore
 
+    def __contains__(self, key: Any) -> bool:
+        return key in self._get_object_with_check()  # type: ignore
+
+    def __iter__(self) -> Any:
+        return self._get_object_with_check().__iter__()  # type: ignore
+
     def __float__(self) -> str:
         # in this case we are looking for the pointed data
         return self.get_object().__float__()  # type: ignore
 
+    def __int__(self) -> int:
+        # in this case we are looking for the pointed data
+        return self.get_object().__int__()  # type: ignore
+
     def __str__(self) -> str:
         # in this case we are looking for the pointed data
         return self.get_object().__str__()
@@ -415,7 +444,7 @@ class IndirectObject(PdfObject):
         self, stream: StreamType, encryption_key: Union[None, str, bytes] = None
     ) -> None:
         if encryption_key is not None:  # deprecated
-            deprecate_no_replacement(
+            deprecation_no_replacement(
                 "the encryption_key parameter of write_to_stream", "5.0.0"
             )
         stream.write(f"{self.idnum} {self.generation} R".encode())
@@ -454,7 +483,7 @@ FLOAT_WRITE_PRECISION = 8  # shall be min 5 digits max, allow user adj
 class FloatObject(float, PdfObject):
     def __new__(
         cls, value: Any = "0.0", context: Optional[Any] = None
-    ) -> "FloatObject":
+    ) -> Self:
         try:
             value = float(value)
             return float.__new__(cls, value)
@@ -484,6 +513,7 @@ class FloatObject(float, PdfObject):
 
         Returns:
             Hash considering type and value.
+
         """
         return hash((self.__class__, self.as_numeric))
 
@@ -491,8 +521,7 @@ class FloatObject(float, PdfObject):
         if self == 0:
             return "0.0"
         nb = FLOAT_WRITE_PRECISION - int(log10(abs(self)))
-        s = f"{self:.{max(1,nb)}f}".rstrip("0").rstrip(".")
-        return s
+        return f"{self:.{max(1, nb)}f}".rstrip("0").rstrip(".")
 
     def __repr__(self) -> str:
         return self.myrepr()  # repr(float(self))
@@ -504,7 +533,7 @@ class FloatObject(float, PdfObject):
         self, stream: StreamType, encryption_key: Union[None, str, bytes] = None
     ) -> None:
         if encryption_key is not None:  # deprecated
-            deprecate_no_replacement(
+            deprecation_no_replacement(
                 "the encryption_key parameter of write_to_stream", "5.0.0"
             )
         stream.write(self.myrepr().encode("utf8"))
@@ -513,7 +542,7 @@ class FloatObject(float, PdfObject):
 class NumberObject(int, PdfObject):
     NumberPattern = re.compile(b"[^+-.0-9]")
 
-    def __new__(cls, value: Any) -> "NumberObject":
+    def __new__(cls, value: Any) -> Self:
         try:
             return int.__new__(cls, int(value))
         except ValueError:
@@ -538,6 +567,7 @@ class NumberObject(int, PdfObject):
 
         Returns:
             Hash considering type and value.
+
         """
         return hash((self.__class__, self.as_numeric()))
 
@@ -548,7 +578,7 @@ class NumberObject(int, PdfObject):
         self, stream: StreamType, encryption_key: Union[None, str, bytes] = None
     ) -> None:
         if encryption_key is not None:  # deprecated
-            deprecate_no_replacement(
+            deprecation_no_replacement(
                 "the encryption_key parameter of write_to_stream", "5.0.0"
             )
         stream.write(repr(self).encode("utf8"))
@@ -556,7 +586,7 @@ class NumberObject(int, PdfObject):
     @staticmethod
     def read_from_stream(stream: StreamType) -> Union["NumberObject", "FloatObject"]:
         num = read_until_regex(stream, NumberObject.NumberPattern)
-        if num.find(b".") != -1:
+        if b"." in num:
             return FloatObject(num)
         return NumberObject(num)
 
@@ -590,6 +620,7 @@ class ByteStringObject(bytes, PdfObject):
 
         Returns:
             Hash considering type and value.
+
         """
         return hash((self.__class__, bytes(self)))
 
@@ -602,13 +633,22 @@ class ByteStringObject(bytes, PdfObject):
         self, stream: StreamType, encryption_key: Union[None, str, bytes] = None
     ) -> None:
         if encryption_key is not None:  # deprecated
-            deprecate_no_replacement(
+            deprecation_no_replacement(
                 "the encryption_key parameter of write_to_stream", "5.0.0"
             )
         stream.write(b"<")
         stream.write(binascii.hexlify(self))
         stream.write(b">")
 
+    def __str__(self) -> str:
+        charset_to_try = ["utf-16", *list(NameObject.CHARSETS)]
+        for enc in charset_to_try:
+            try:
+                return self.decode(enc)
+            except UnicodeDecodeError:
+                pass
+        raise PdfReadError("Cannot decode ByteStringObject.")
+
 
 class TextStringObject(str, PdfObject):  # noqa: SLOT000
     """
@@ -624,37 +664,40 @@ class TextStringObject(str, PdfObject):  # noqa: SLOT000
     utf16_bom: bytes
     _original_bytes: Optional[bytes] = None
 
-    def __new__(cls, value: Any) -> "TextStringObject":
-        org = None
+    def __new__(cls, value: Any) -> Self:
+        original_bytes = None
         if isinstance(value, bytes):
-            org = value
+            original_bytes = value
             value = value.decode("charmap")
-        o = str.__new__(cls, value)
-        o._original_bytes = org
-        o.autodetect_utf16 = False
-        o.autodetect_pdfdocencoding = False
-        o.utf16_bom = b""
-        if value.startswith(("\xfe\xff", "\xff\xfe")):
-            assert org is not None  # for mypy
+        text_string_object = str.__new__(cls, value)
+        text_string_object._original_bytes = original_bytes
+        text_string_object.autodetect_utf16 = False
+        text_string_object.autodetect_pdfdocencoding = False
+        text_string_object.utf16_bom = b""
+        if original_bytes is not None and original_bytes[:2] in {codecs.BOM_UTF16_LE, codecs.BOM_UTF16_BE}:
+            # The value of `original_bytes` is only set for inputs being `bytes`.
+            # If this is UTF-16 data according to the BOM (first two characters),
+            # perform special handling. All other cases should not need any special conversion
+            # due to already being a string.
             try:
-                o = str.__new__(cls, org.decode("utf-16"))
-            except UnicodeDecodeError as exc:
+                text_string_object = str.__new__(cls, original_bytes.decode("utf-16"))
+            except UnicodeDecodeError as exception:
                 logger_warning(
-                    f"{exc!s}\ninitial string:{exc.object!r}",
+                    f"{exception!s}\ninitial string:{exception.object!r}",
                     __name__,
                 )
-                o = str.__new__(cls, exc.object[: exc.start].decode("utf-16"))
-            o._original_bytes = org
-            o.autodetect_utf16 = True
-            o.utf16_bom = org[:2]
+                text_string_object = str.__new__(cls, exception.object[: exception.start].decode("utf-16"))
+            text_string_object._original_bytes = original_bytes
+            text_string_object.autodetect_utf16 = True
+            text_string_object.utf16_bom = original_bytes[:2]
         else:
             try:
-                encode_pdfdocencoding(o)
-                o.autodetect_pdfdocencoding = True
+                encode_pdfdocencoding(text_string_object)
+                text_string_object.autodetect_pdfdocencoding = True
             except UnicodeEncodeError:
-                o.autodetect_utf16 = True
-                o.utf16_bom = codecs.BOM_UTF16_BE
-        return o
+                text_string_object.autodetect_utf16 = True
+                text_string_object.utf16_bom = codecs.BOM_UTF16_BE
+        return text_string_object
 
     def clone(
         self,
@@ -678,6 +721,7 @@ class TextStringObject(str, PdfObject):  # noqa: SLOT000
 
         Returns:
             Hash considering type and value.
+
         """
         return hash((self.__class__, self.original_bytes))
 
@@ -691,8 +735,7 @@ class TextStringObject(str, PdfObject):  # noqa: SLOT000
         """
         if self._original_bytes is not None:
             return self._original_bytes
-        else:
-            return self.get_original_bytes()
+        return self.get_original_bytes()
 
     def get_original_bytes(self) -> bytes:
         # We're a text string object, but the library is trying to get our raw
@@ -703,14 +746,12 @@ class TextStringObject(str, PdfObject):  # noqa: SLOT000
         if self.autodetect_utf16:
             if self.utf16_bom == codecs.BOM_UTF16_LE:
                 return codecs.BOM_UTF16_LE + self.encode("utf-16le")
-            elif self.utf16_bom == codecs.BOM_UTF16_BE:
+            if self.utf16_bom == codecs.BOM_UTF16_BE:
                 return codecs.BOM_UTF16_BE + self.encode("utf-16be")
-            else:
-                return self.encode("utf-16be")
-        elif self.autodetect_pdfdocencoding:
+            return self.encode("utf-16be")
+        if self.autodetect_pdfdocencoding:
             return encode_pdfdocencoding(self)
-        else:
-            raise Exception("no information about original bytes")  # pragma: no cover
+        raise Exception("no information about original bytes")  # pragma: no cover
 
     def get_encoded_bytes(self) -> bytes:
         # Try to write the string out as a PDFDocEncoding encoded string. It's
@@ -735,7 +776,7 @@ class TextStringObject(str, PdfObject):  # noqa: SLOT000
         self, stream: StreamType, encryption_key: Union[None, str, bytes] = None
     ) -> None:
         if encryption_key is not None:  # deprecated
-            deprecate_no_replacement(
+            deprecation_no_replacement(
                 "the encryption_key parameter of write_to_stream", "5.0.0"
             )
         bytearr = self.get_encoded_bytes()
@@ -755,13 +796,9 @@ class TextStringObject(str, PdfObject):  # noqa: SLOT000
 
 class NameObject(str, PdfObject):  # noqa: SLOT000
     delimiter_pattern = re.compile(rb"\s+|[\(\)<>\[\]{}/%]")
-    surfix = b"/"
-    renumber_table: ClassVar[Dict[str, bytes]] = {
-        "#": b"#23",
-        "(": b"#28",
-        ")": b"#29",
-        "/": b"#2F",
-        "%": b"#25",
+    prefix = b"/"
+    renumber_table: ClassVar[dict[str, bytes]] = {
+        **{chr(i): f"#{i:02X}".encode() for i in b"#()<>[]{}/%"},
         **{chr(i): f"#{i:02X}".encode() for i in range(33)},
     }
 
@@ -783,6 +820,7 @@ class NameObject(str, PdfObject):  # noqa: SLOT000
 
         Returns:
             Hash considering type and value.
+
         """
         return hash((self.__class__, self))
 
@@ -790,7 +828,7 @@ class NameObject(str, PdfObject):  # noqa: SLOT000
         self, stream: StreamType, encryption_key: Union[None, str, bytes] = None
     ) -> None:
         if encryption_key is not None:  # deprecated
-            deprecate_no_replacement(
+            deprecation_no_replacement(
                 "the encryption_key parameter of write_to_stream", "5.0.0"
             )
         stream.write(self.renumber())
@@ -798,9 +836,9 @@ class NameObject(str, PdfObject):  # noqa: SLOT000
     def renumber(self) -> bytes:
         out = self[0].encode("utf-8")
         if out != b"/":
-            deprecate_no_replacement(
+            deprecation_no_replacement(
                 f"Incorrect first char in NameObject, should start with '/': ({self})",
-                "6.0.0",
+                "5.0.0",
             )
         for c in self[1:]:
             if c > "~":
@@ -813,6 +851,26 @@ class NameObject(str, PdfObject):  # noqa: SLOT000
                     out += c.encode("utf-8")
         return out
 
+    def _sanitize(self) -> "NameObject":
+        """
+        Sanitize the NameObject's name to be a valid PDF name part
+        (alphanumeric, underscore, hyphen). The _sanitize method replaces
+        spaces and any non-alphanumeric/non-underscore/non-hyphen with
+        underscores.
+
+        Returns:
+            NameObject with sanitized name.
+        """
+        name = str(self).removeprefix("/")
+        name = re.sub(r"\ ", "_", name)
+        name = re.sub(r"[^a-zA-Z0-9_-]", "_", name)
+        return NameObject("/" + name)
+
+    @classproperty
+    def surfix(cls) -> bytes:  # noqa: N805
+        deprecation_with_replacement("surfix", "prefix", "5.0.0")
+        return b"/"
+
     @staticmethod
     def unnumber(sin: bytes) -> bytes:
         i = sin.find(b"#", 0)
@@ -831,8 +889,8 @@ class NameObject(str, PdfObject):  # noqa: SLOT000
     @staticmethod
     def read_from_stream(stream: StreamType, pdf: Any) -> "NameObject":  # PdfReader
         name = stream.read(1)
-        if name != NameObject.surfix:
-            raise PdfReadError("name read error")
+        if name != NameObject.prefix:
+            raise PdfReadError("Name read error")
         name += read_until_regex(stream, NameObject.delimiter_pattern)
         try:
             # Name objects should represent irregular characters
@@ -853,11 +911,10 @@ class NameObject(str, PdfObject):  # noqa: SLOT000
                     __name__,
                 )
                 return NameObject(name.decode("charmap"))
-            else:
-                raise PdfReadError(
-                    f"Illegal character in NameObject ({name!r}). "
-                    "You may need to adjust NameObject.CHARSETS.",
-                ) from e
+            raise PdfReadError(
+                f"Illegal character in NameObject ({name!r}). "
+                "You may need to adjust NameObject.CHARSETS.",
+            ) from e
 
 
 def encode_pdfdocencoding(unicode_string: str) -> bytes:
@@ -877,6 +934,7 @@ def is_null_or_none(x: Any) -> TypeGuard[Union[None, NullObject, IndirectObject]
     """
     Returns:
         True if x is None or NullObject.
+
     """
     return x is None or (
         isinstance(x, PdfObject)
